//*****************************************************************************
//*
//*   File :  bwstal.c
//*
//*   Owner:  Brad Stumpf
//*           406 Cozby Ave.
//*           Coppell, TX  75019
//*                  
//*   EMAIL:  Internet -> BSTUMPF@AOL.COM
//*           AOL -> BStumpf
//*           Premia BBS Id -> Brad Stumpf
//*
//*   Desc :  Windows DLL for use with CodeWright V2.0f or higher to implement
//*           Tandem Computers TAL language extention processing
//*           (Chromacoding and indentation/template completion logic).
//*
//*           ChromaCoding supports TAL keywords, compiler command-lines (ie.
//*           ?xxxx), comment lines (ie. ! and --) and G90 procedure names.
//*           Each of these will be in a different color (if configured).
//*
//*   Codewright:
//*       Premia Corporation
//*       1075 NW Murray Blvd., Suite 268
//*       Portland OR 97229
//*
//*****************************************************************************

#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"
#include "tal.h"


LIBMAIN

#define TAL_VERSION  "V3.0g-1.0j (30APR94)"

#define MAX_BUF_LEN        132

static LPVOID _tal_hash_names        = NIL;  /* pointer to the hash table containing */
                                             /* all 'TAL' keyword names. */

static LPVOID _tal_hash_templates    = NIL;  /* pointer to the hash table containing */
                                             /* an abbreviated name and its template */

static LPVOID _tal_hash_names_1      = NIL;  /* pointer to the hash table containing */
                                           /* user defined keyword names. */
static int _tal_hash_size_1          = 47;
static LPVOID _tal_hash_names_2      = NIL;  /* pointer to the hash table containing */
                                           /* user defined keyword names. */
static int _tal_hash_size_2          = 47;
 
static long _tal_in_comment( long offset );

void DLL _tal_init( void );
BOOL DLL _tal_expand( void );
void DLL _tal_language_colors_update( long firstline, long lastline );
void DLL _tal_language_colors( long numlines );
void DLL _tal_assign_template( LPSTR abbrev, LPSTR template );
void DLL _tal_add_keyword( LPSTR keyword, int table );
BOOL DLL _tal_indent( void );

// -----------------------------------------------------------------
LPSTR DLL 
_tal_version( void )
{
   static LPSTR ver = NIL; 

   if ( !ver )
   {
      ver = StrNew( TAL_VERSION ); 
   }

   return ver;
}

void DLL 
_init()
{
   _tal_init();
}

/*
 ** _tal_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_tal_init( void )
{
   static BOOL firstTime = TRUE;
   int    i;

   if (firstTime)
   {
      LibExport( "LPSTR _tal_version");
      LibExport( "void _tal_init");
      LibExport( "BOOL _tal_expand");
      LibExport( "_tal_assign_template LPSTR LPSTR");
      LibExport( "_tal_add_keyword LPSTR int");
      LibExport( "BOOL _tal_indent( void ) ");
      LibExport( "_tal_language_colors_update long long");
      LibExport( "_tal_language_colors long");
      LibExport( "_tal_routines" );
      LibExport( "_tal_hash_size int int" );

/*------------ Pre-Premia support for these functions (30APR94-BWS) ---------*/
      if (LibFunctionExists( "ExtAddUserKeyword" ) == 0)
         LibExport( "cwstart:ExtAddUserKeyword LPSTR LPSTR int" );
      if (LibFunctionExists( "ExtUserTableSize" ) == 0)
         LibExport( "cwstart:ExtUserTableSize LPSTR int int" );
      if (LibFunctionExists( "ExtReadUserKeywordFile" ) == 0)
         LibExport( "BOOL cwstart:ExtReadUserKeywordFile LPSTR LPSTR int" );
/*------------ Pre-Premia support for these functions (30APR94-BWS) ---------*/

      firstTime = FALSE;

      /* Create a hash table to hold all of the 'TAL' keyword names.
       */
      _tal_hash_names = HashCreateTable( _tal_names_count,
                                         sizeof( LPSTR ),
                                         HASH_IGN_CASE );
      i = 0;

      while ( _tal_names[ i ] != NIL )
      {
         LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _tal_hash_names, 
                        _tal_names[i], strlen( _tal_names[i] ) );
         if (lpf)
         {
            *lpf = StrNew( _tal_names[i] );
            i++;
         }
      }

      /* Create a hash table to hold all of the 'TAL' template names.
       */
      _tal_hash_templates = HashCreateTable( _tal_templates_count,
                                             sizeof( LPSTR ),
                                             HASH_IGN_CASE );
      i = 0;

      while ( _tal_templates[ i ].abbrev != NIL )
      {
         _tal_assign_template( _tal_templates[i].abbrev, _tal_templates[i].template );
         i++;
      }
   }
}


void DLL
_tal_hash_size( int ht, int size )
{
   if (size > 0)
   {
      if (ht == 1)
         _tal_hash_size_1 = size;
      else if (ht == 2)
         _tal_hash_size_2 = size;
   }
}


/*
 ** _tal_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'TAL' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_tal_expand( void )
{
   LPVOID htEntry;
   LPSTR  _tal_string;
   LPSTR  str;
   BOOL   insmode = BufQInsertMode();

   MarkSavePos();

   /* Check to see if we are currently positioned inside a 'TAL' comment
    */
   if (_tal_in_comment( -1L ) != -1)
      goto _tal_expand_incomplete;

   str = BufReadStr( 65535 );     /* read entire line */

   _tal_string = StrTrim( str, " \t" );
   StrFree( str );

   if (!_tal_string)
      goto _tal_expand_incomplete;

   /* Search for a keyword match in the hash table.
    */
   htEntry = HashFindEntry( _tal_hash_templates, _tal_string, strlen( _tal_string ) );
   StrFree( _tal_string );

   if (htEntry)
   {
      /* A template exists for the current keyword, delete the keyword
       * and insert the template.
       */
      MarkDropPos( 1 );
      str = *(LPSTR*)htEntry;
      BufDelToEOL();

      BufSetInsertMode( TRUE );
      _ext_expand_template( str );
      BufSetInsertMode( insmode );
   } 
   else
      goto _tal_expand_incomplete;
   return TRUE;

_tal_expand_incomplete:
   MarkRestorePos();

   return FALSE;
}



/*
 ** _tal_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
void DLL
_tal_language_colors_update( long firstline, long lastline )
{
   long startoff;

   if (firstline && lastline)
   {
      /* Begin coloring at the previous comment if
       * currently located in the middle of one.
       */
      _PosInit( 0 );
      if (firstline == 1)
         _PosPrevLine( 0L );
      else
         _PosNextLine( firstline - 1);
      startoff = _PosQOffset();

      /* Clear all attributes for the line range and
       * the reset them within _cob_language_colors.
       */
      AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
      _PosInit( startoff );
      _tal_language_colors( lastline - firstline + 1 );
   }
}


/*
 ** _tal_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_tal_language_colors( long numlines )
{
   long  startoff;
   long  startline;
   long  endoff;
   UINT  ch;
   LPSTR buf;
   int   i = 0;
   int   comment_color = ColorComments( -1 );
   int   keyword_color = ColorKeywords( -1 );
   int   quote_color   = ColorAlternate1( -1 );
   int   number_color  = ColorAlternate2( -1 );
   int   bracket_color = ColorAlternate3( -1 );
   int   preproc_color = ColorAlternate4( -1 );
   int   G90proc_color = ColorAlternate5( -1 );
   int   altkw_color   = ColorAlternate6( -1 );
   int   oper_color    = ColorAlternate7( -1 );



   buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

   /* Start at the top of the buffer or current position 
    */
   if (numlines == -1)
   {
      _PosInit( 0 );  
      numlines = 0x7FFFFFFF;
   }
   else
   {
      startoff  = _PosQOffset();
      startline = _PosQLine();
      endoff    = _tal_in_comment( startoff );
      if (endoff == -1)
         _PosInit( startoff );
      else
      {
         _PosInit( endoff );
         numlines += (startline - _PosQLine());
         _PosInit( endoff );
      }
   }
   ch = _PosCurrentChar();
   startoff = _PosQOffset();

   while ((ch != EOF_CHAR) && (numlines > 0))
   {
      if (ch == '\n')
         numlines--;

      if (ch == '?')
      {
         startoff = _PosQOffset();
         _PosPrevLine( 0L );
         if ( startoff == _PosQOffset() )
         {
            _PosNextLine( 0L );
            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( preproc_color, endoff - startoff );
            _PosInit( endoff );
            ch = _PosCurrentChar();
            continue;
         }
         _PosInit( startoff );
      }

      if (ch == '"')
      {
         UINT quote = ch;

         startoff = _PosQOffset();
         i = 0;  // reset for keyword/proc Hash lookup logic below

         do 
         {
            ch = _PosNextChar();
            if ((ch == EOF_CHAR) || (ch == '\n'))
               break;
         } while (ch != quote);

         if (ch == quote)
            ch = _PosNextChar();

         /* color the string */
         if (quote_color)
         {
            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( quote_color, endoff - startoff );
            _PosInit( endoff );
         }

         continue;
      }
      else if ( number_color
             && ( isdigit( ch ) 
               || ch == '%' ) )
      {
         UINT lastch  = _PosPrevChar();
         long skipcnt = 0L;
         int  numflag = FALSE;
         int  hexflag = FALSE;

         if ( ch == '%'
           || (lastch != EOF_CHAR
             &&lastch != '-' ) )
            _PosNextChar();

         if ( ch == '%' )
         {
            ch = _PosNextChar();
            if ( ch != EOF_CHAR
              && ( _fstrchr("HhBb", ch) != 0 ) )
            {
               if ( ch == 'H' || ch == 'h' )
                  hexflag = TRUE;

               ch = _PosNextChar();
               if ( isdigit( ch )
                 || ( hexflag && isxdigit( ch ) ) )
                  skipcnt = 2L;
               else
                  ch = _PosPrevChar();
            }
            else
               skipcnt = 1L;
         }

         startoff = _PosQOffset() - skipcnt;

         while ( isdigit( ch )
              || ( hexflag && isxdigit( ch ) ) )
         {
            numflag = TRUE;
            ch = _PosNextChar();
         }

         if ( !numflag
           || ( lastch != EOF_CHAR
             && ( _fstrchr(" .,[('=:<>+-*/\\\n\r\t", lastch) == 0 ) ) )
            continue;

         if ( ch != EOF_CHAR
           && ( _fstrchr("DdFf", ch) != 0 ) )
            ch = _PosNextChar();

         endoff = _PosQOffset();
         _PosInit( startoff );
         _PosSetColor( number_color, endoff - startoff );
         _PosInit( endoff );

         continue;
      }
      else if ( _TAL_VALID_WORD_CH( ch ) )
      {
         UINT lastch = _PosPrevChar();

         if (lastch != EOF_CHAR)
            _PosNextChar();

         i = 0;
         startoff = _PosQOffset();

         do
         {
            if (i < MAX_BUF_LEN)
               buf[i++] = (char)ch;
            ch = _PosNextChar();
         } while ( _TAL_VALID_WORD_CH( ch )
                 );

         if ((lastch != ' ' 
                  && lastch != '\t' 
                  && lastch != '\r' 
                  && lastch != '\n' 
                  && lastch != EOF_CHAR
                  && !ispunct( lastch ))
               || _TAL_VALID_WORD_PUNCT( ch )
            )
            i = 0;
         continue;
      }
      else if (i)
      {
         int color = 0;

         if (_tal_hash_names_1 && HashFindEntry( _tal_hash_names_1, buf, i ))
            color = G90proc_color;
         else if (_tal_hash_names_2 && HashFindEntry( _tal_hash_names_2, buf, i ))
            color = altkw_color;
         else if (HashFindEntry( _tal_hash_names, buf, i ))
            color = keyword_color;

         if (color)
         {
            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( color, endoff - startoff );
            _PosInit( endoff );
         }

         i = 0;
      } 

      if ( (ch == '-') || (ch == '!') )
      {
         startoff = _PosQOffset();

         if (ch == '-')
         {
            ch = _PosNextChar();

            if ( ch == '-')
            {
               _PosNextLine( 0 );
               endoff = _PosQOffset();
               _PosInit( startoff );
               _PosSetColor( comment_color, endoff - startoff );
               _PosInit( endoff );
               ch = _PosCurrentChar();
               continue;
            }
            else
               ch = _PosPrevChar();
         }
         else
         {
            ch = _PosNextChar();
            while ((ch != EOF_CHAR) && (ch != '\n') && (ch != '!'))
                  ch = _PosNextChar();

            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( comment_color, endoff - startoff + 
                          ( ch == '!' ? 1L : 0L ) );
            _PosInit( endoff );

            if (ch == '!')
               ch = _PosNextChar();
            continue;
         }
      }
      else if (bracket_color && ( _fstrchr("()", ch) != 0) )
      {                                        
         long coff;                            

         coff = _PosQOffset();            
         _PosInit(coff);                  
         _PosSetColor(bracket_color, 1);    
         _PosInit(coff);                  
      }                                   
      else if (keyword_color && ( _fstrchr("[]@", ch) != 0) )
      {                                        
         long coff;                            

         coff = _PosQOffset();            
         _PosInit(coff);                  
         _PosSetColor(keyword_color, 1);    
         _PosInit(coff);                  
      }                                   

      if ( ( oper_color  ||  keyword_color )
        && ( _fstrchr("'=:<>+-*/\\", ch) != 0) )
      {                                        
         UINT npch;
         int  xcolor = oper_color;

         startoff = _PosQOffset();              

         switch ( ch )
         {
         case ':':
              npch = _PosNextChar();

              if ( npch == 39 )
              {
                 _PosInit(startoff);
                 npch = _PosPrevChar();
                 if ( npch == '=' )         //reverse move...highlight!
                 {
                    xcolor = keyword_color;
                    startoff = _PosQOffset();
                    _PosNextChar();
                 } 
                 else
                   _PosInit(startoff);
              } 
              else 
              {
                if ( npch != '=' )
                   xcolor = 0;
                _PosInit(startoff);
              }
              break;

         case '<':
              do
              {
                npch = _PosNextChar();
              } while ( npch != EOF_CHAR
                     && ( _fstrchr( " \t\r", npch ) != 0 ) );

              if ( isdigit( (BYTE)npch ) )
              {
                 do
                 {
                   npch = _PosNextChar();
                 } while ( npch != EOF_CHAR
                        && ( isdigit( (BYTE)npch )
                          || ( _fstrchr( " \t\r", npch ) != 0 ) ) );

                 if ( npch == ':'
                   || npch == '>' )
                    xcolor = keyword_color;
              }
              _PosInit(startoff);
              break;

         case '>':
              do
              {
                npch = _PosPrevChar();
              } while ( npch != EOF_CHAR
                     && ( _fstrchr( " \t\r", npch ) != 0 ) );

              if ( isdigit( (BYTE)npch ) )
              {
                 do
                 {
                   npch = _PosPrevChar();
                 } while ( npch != EOF_CHAR
                        && ( isdigit( (BYTE)npch )
                          || ( _fstrchr( " \t\r", npch ) != 0 ) ) );

                 if ( npch == ':'
                   || npch == '<' )
                    xcolor = keyword_color;
              }
              _PosInit(startoff);
              break;

         case '-':
              npch = _PosNextChar();
              if ( npch == '>' )
                  xcolor = keyword_color;
              else
                  _PosInit(startoff);
              break;

         case '*':
              do
              {
                npch = _PosPrevChar();
              } while ( npch != EOF_CHAR
                     && ( _fstrchr( " \t\r", npch ) != 0 ) );

              if ( npch == '(' )
                  xcolor = keyword_color;
              _PosInit(startoff);
              break;

         default:
              break;
         }

         if ( xcolor )
         {
            endoff = _PosQOffset();
            _PosInit(startoff);                  
            _PosSetColor( xcolor, endoff - startoff + 1 );
            _PosInit(endoff);                 
         }
      }                                   

      ch = _PosNextChar();
   }
   MemFree( buf );
}


void DLL
_tal_add_keyword( LPSTR keyword, int table )
{
   if (keyword && *keyword && table >= 0)
   {
      LPSTR XFAR *lpf;
      LPVOID  ht;

      switch( table )
      {
      case 0:
         ht = _tal_hash_names;
         break;
      case 1:
         if (_tal_hash_names_1 == NULL)
            _tal_hash_names_1 = HashCreateTable( _tal_hash_size_1,
                                                 sizeof( LPSTR ), 
                                                 HASH_IGN_CASE );
         ht = _tal_hash_names_1;
         break;
      case 2:
         if (_tal_hash_names_2 == NULL)
            _tal_hash_names_2 = HashCreateTable( _tal_hash_size_2,
                                                 sizeof( LPSTR ),
                                                 HASH_IGN_CASE );
         ht = _tal_hash_names_2;
         break;
      }

      lpf = (LPSTR XFAR *)HashGetEntry( ht, keyword, strlen( keyword ) );
      if (lpf)
         *lpf = StrNew( keyword );
   }
}


/*
 ** _tal_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'TAL' template strings
 *    without changing the code.
 */
void DLL
_tal_assign_template( LPSTR abbrev, LPSTR template )
{
   
   _ext_assign_template( _tal_hash_templates, abbrev, template );

}


/*
 ** _tal_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'TAL' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in C_INDENT_STR - perform smart
 *       indentation.  This may include skipping the next line if
 *       it consists only of an open brace.
 */
BOOL DLL
_tal_indent( void )
{
   long   curline;
   long   col;
   LPSTR  _tal_string;
   LPSTR  str;
   long   matchpos;
//<<   UINT   endch;
   BOOL   insmode = BufQInsertMode();

   MarkSavePos();

   curline = BufQCurrentLine();

   /* Check to see if we are currently positioned inside a 'TAL' comment
    */
   if (_tal_in_comment( -1L ) != -1)
      goto _tal_indent_incomplete;

   /* Not in a comment so indent according to the current 'TAL' construct.
    */
   if (!SrchFind( "^[ \t]*[a-zA-Z0-9_^$]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
         || (BufQCurrentLine() != curline) )
      goto _tal_indent_incomplete;

   /* Read the keyword at the beginning of the current line.
    */
   col = BufQCurrentCol();
   SrchFind( "(^|[^a-zA-Z0-9_^$])\\c[a-zA-Z0-9_^$]", SEARCH_REGEX , NIL );
   str = BufReadStr( col - BufQCurrentCol() );
   _tal_string = StrTrim( str, " \t" );
   StrFree( str );

   /* Set prefix and suffix to \377 to perform a quick lookup of the name
    */
   str = StrNew( "\377" );
   _tal_string = StrAppend( _tal_string, str );
   str = StrAppend( str, _tal_string );
   StrFree( _tal_string );
   
   matchpos = StrMatch( str, C_INDENT_STR, SEARCH_FORWARD, NIL );
   StrFree( str );

   if (!matchpos)
      /* Not a 'TAL' keyword, do normal indentation. */
      goto _tal_indent_incomplete;
      
   MarkRestorePos();
   MarkSavePos();

   /* Check to see if a 'begin' is by itself on the next line. If so, skip
    * it an indent on the line following it.
    */
   if (SrchFind( "^[ \t]*begin\\c[ \t]*$",
                 SEARCH_FORWARD | SEARCH_REGEX | SEARCH_IGCASE, NIL )
         && (BufQCurrentLine() == curline + 1) )
      MarkDropPos( 1 );
   else
      MarkRestorePos();

//<<_tal_indent_next_line:
   BufSetInsertMode( TRUE );
   _ext_insert_indented_EOL();
   BufInsertChar( '\t' );
   BufSetInsertMode( insmode );
   return TRUE;

_tal_indent_incomplete:
   MarkRestorePos();

   return FALSE;
}



/*--  Private functions --*/


/*
 ** _tal_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'TAL' comment by searching from the beginning of
 *  line specfiec in offset for a ! or -- comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_tal_in_comment( long offset )
{
   UINT  ch;
   int   sccnt = 0;
   long  scoff;

   /* Check to see if we are currently positioned inside a 'TAL' comment
    */
   _PosInit( offset );
   _PosPrevLine( 0L );
   ch = _PosCurrentChar();

   while ( (ch != EOF_CHAR) && (ch != '\n' )
        && (_PosQOffset() <= offset) )
   {
      
      if ( ch == '!' )
      {
         if ( (++sccnt & 1) )
            scoff = _PosQOffset();
      }
      else if ( ch == '-' )
      {
         ch = _PosNextChar();
         if (ch == '-')
            return( _PosQOffset() - 1L );  /* currrent line is in a comment */
                                  // BWS-04JAN94
         continue;
      }
      else if ( ch == '"' )
      {
         ch = _PosNextChar();
         while (ch != EOF_CHAR && ch != '"' && ch != '\n')
            ch = _PosNextChar();

         if ( ch != '"')
            continue;
      }

      ch = _PosNextChar();
   }

   return ( (sccnt & 1) ? scoff : -1L );
}



void DLL
_tal_routines( void )
{
   int   matches = 0;
   LPSTR str     = "^[ \t]*((int[0-9()]*|fixed)[ \t\n]+proc|proc)[ \t\n]+[a-z0-9_^]+[ \t\n]*(\\(|[ a-z\t\n]*;)";
   long  curline;
//<<   UINT  ch;
   long  lines;

   SysWaitStart();

   MarkSavePos();

   BufSetAscii();
   MovTopBuf();

   while (SrchFind( str, SEARCH_REGEX | SEARCH_FORWARD | SEARCH_IGCASE, 0 ))
   {
      curline = BufQCurrentLine();

      if ( SrchFind( ";", SEARCH_FORWARD, 0 ) )
      {
         if (!matches)
            AttrSetVisible( 0, 0, 0 );
         matches++;

         lines = BufQCurrentLine() - curline;
         AttrSetVisible( curline, curline + (lines < 0L ? 0L : lines), 1 );
         MovHome();
      }

      if (!MovDown( 1 ))
         break; 
   }

   if (matches)
   {
      AttrSetVisible( 1, 1, 1 );  //>>BWS  Always make line #1 visible
      BufSetCompact();
      MarkDropPos( 1 );
      MovTopBuf();
   }
   else
      MarkRestorePos();

   SysWaitEnd();

   if ( matches )
      MsgPrintf( MESSAGE_MSG_LEVEL, "%d procedure%s found.", matches,
                 ( matches > 1 ? "s" : "" ) );
   else
      MsgWarning( "No procedures found." );
}
