//*****************************************************************************
//*
//*   File :  bwstal.h
//*
//*   Owner:  Brad Stumpf
//*           406 Cozby Ave.
//*           Coppell, TX  75019
//*                  
//*   EMAIL:  Internet -> BSTUMPF @AOL.COM
//*           AOL -> BStumpf
//*           Premia BBS Id -> Brad Stumpf
//*
//*   Desc :  Header file for BWSTAL.C.
//*
//*   Codewright:
//*       Premia Corporation
//*       1075 NW Murray Blvd., Suite 268
//*       Portland OR 97229
//*
//*****************************************************************************


/* define for termining if character should be considered for
   'word' processing
 */
#define _TAL_VALID_WORD_PUNCT( X )  ( (X == '^') || (X == '.') || (X == '$') || (X == '_') )
#define _TAL_VALID_WORD_CH( X )     ( isalnum( X ) || _TAL_VALID_WORD_PUNCT( X ) )


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR C_INDENT_STR = "\377if\377while\377else\377do\377for\377begin\377end\377";


/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _tal_hash_templates.
 */
typedef struct _LanguageTemplate {         
   LPSTR abbrev;
   LPSTR template;
} LanguageTemplate;


/*
 * _tal_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _tal_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
#define _tal_templates_count  15
LanguageTemplate _tal_templates[] = {
   { "_",                  "--\n--  &\n--" },
   { "IF",                 "IF  &  THEN\n" },
   { "IFF",                "IF  &  THEN\n  BEGIN\n\nEND;" },
   { "IFE",                "IF  &  THEN\n  BEGIN\n\nEND\n@@ELSE\n  BEGIN\n\nEND;" },
   { "WH",                 "WHILE  &  DO\n" },
   { "WHI|WHIL|WHILE",     "WHILE  &  DO\nBEGIN\n\nEND;" },
   { "DO",                 "DO\nBEGIN\n&\nEND UNTIL  ;" },
   { "FO|FOR",             "FOR  & :=   TO   DO\nBEGIN\n\nEND;" },
   { "FORD",               "FOR  & :=   DOWNTO   DO\nBEGIN\n\nEND;" },
   { "CA|CAS|CASE",        "CASE  &  OF\nBEGIN\n\n  OTHERWISE\n@@END;" },
   { "BE|BEG|BEGI|BEGIN",  "BEGIN\n  &\n@@END;" },
   { "EL",                 "ELSE\n  &" },
   { "ELS|ELSE",           "ELSE\n  BEGIN\n  &\n@@END;" },
   { "RE",                 "RETURN;" },
   { "RET|RETU|RETUR",     "RETURN " },
   { NIL,                  NIL },
};


/* Define the list of 'TAL' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
#define _tal_names_count  106
LPSTR _tal_names[] = {
   "$ABS"
  ,"$ALPHA"
  ,"$BITLENGTH"
  ,"$BITOFFSET"
  ,"$CARRY"
  ,"$COMP"
  ,"$DBL"
  ,"$DBLL"
  ,"$DBLR"
  ,"$DFIX"
  ,"$EFLT"
  ,"$EFLTR"
  ,"$FIX"
  ,"$FIXD"
  ,"$FIXI"
  ,"$FIXL"
  ,"$FIXR"
  ,"$FLT"
  ,"$FLTR"
  ,"$HIGH"
  ,"$IFIX"
  ,"$INT"
  ,"$INTR"
  ,"$LADR"
  ,"$LEN"
  ,"$LFIX"
  ,"$LMAX"
  ,"$LMIN"
  ,"$MAX"
  ,"$MIN"
  ,"$NUMERIC"
  ,"$OCCURS"
  ,"$OFFSET"
  ,"$OVERFLOW"
  ,"$PARAM"
  ,"$POINT"
  ,"$READCLOCK"
  ,"$RP"
  ,"$SCALE"
  ,"$SPECIAL"
  ,"$TYPE"
  ,"$UDBL"
  ,"$USERCODE"
  ,"$XADR"
  ,".EXT"
  ,".SG"
  ,"AND"
  ,"ASSERT"
  ,"AT"
  ,"BEGIN"
  ,"BELOW"
  ,"BIT_FILLER"
  ,"BLOCK"
  ,"BY"
  ,"CALL"
  ,"CALLABLE"
  ,"CASE"
  ,"CODE"
  ,"DEFINE"
  ,"DO"
  ,"DOWNTO"
  ,"DROP"
  ,"ELSE"
  ,"END"
  ,"ENTRY"
  ,"EXTENSIBLE"
  ,"EXTERNAL"
  ,"FILLER"
  ,"FIXED"
  ,"FOR"
  ,"FORWARD"
  ,"GOTO"
  ,"IF"
  ,"INT"
  ,"INTERRUPT"
  ,"LABEL"
  ,"LAND"
  ,"LITERAL"
  ,"LOR"
  ,"MAIN"
  ,"NAME"
  ,"NOT"
  ,"OF"
  ,"OR"
  ,"OTHERWISE"
  ,"PRIV"
  ,"PRIVATE"
  ,"PROC"
  ,"REAL"
  ,"RESIDENT"  
  ,"RETURN"
  ,"RSCAN"
  ,"SCAN"
  ,"STACK"
  ,"STORE"
  ,"STRING"
  ,"STRUCT"
  ,"SUBPROC"
  ,"THEN"
  ,"TO"
  ,"UNSIGNED"
  ,"UNTIL"
  ,"USE"
  ,"VARIABLE"
  ,"WHILE"
  ,"XOR"

  ,NIL
};
